#!/usr/bin/perl
# Written by Craig Maloney
# (c) 2005 Craig Maloney
# Released under the GPL

$version_number = "1.3.2";
use Getopt::Long;

$wordfile = "/usr/share/dict/words";
$num_of_words = 50;
$gridsize = 40;
$directions = 8;
$max_iterations = 5; 	# Maximum number of times to try before using drastic measures
$debug = 0;
$nogrow = 0;
$finished = 0; 	# Main loop toggle

@wordlist = ();
@selected_words = ();
@x_offset = (1,0,0,-1,1,1,-1,-1); # R,D,U,L,UR,DR,DL,UL
@y_offset = (0,1,-1,0,-1,1,1,-1); # R,D,U,L,UR,DR,DL,UL

@randomletters = split //, "ABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ";

GetOptions ("size=i" => \$gridsize,
	"directions=i" => \$directions,
	"words=i" => \$num_of_words,
	"all" => \$all,
	"wordfile=s" => \$wordfile,
	"debug" => \$debug,
	"nosolution" => \$nosolution,
	"nogrow" => \$nogrow,
	"version" => \$version,
	"quick" => \$quick,
	"thorough" => \$thorough,
	"help" => \$help );

if ($help) {
	&usage;
	exit(0);
}

if ($version) {
	print "$0 version: $version_number\n";
	exit(0);
}

if ($quick) {
	$max_iterations = 1;
}

if ($thorough) {
	$max_iterations = 20;
}

open WORDLIST, "$wordfile" || die "Can't open wordlist $wordfile\n";
foreach $i (<WORDLIST>) {
	chomp $i;
	push @wordlist, uc $i if ($i ne '' && $i ne ' ');
}
close WORDLIST;

if (scalar @wordlist < $num_of_words) {
	$all = 1;
}

if ($all) {
	foreach $word (@wordlist) {
		push @selected_words, $word;
	}
	$num_of_words = scalar @selected_words;
} else {
	$i=0;
	while ($i<$num_of_words) { 
		$rand_element = int(rand $#wordlist);
		if (length($wordlist[$rand_element]) > 5) {
			push @selected_words, $wordlist[$rand_element];
			print "$wordlist[$rand_element]\n" if $debug;
			$i++;
		}
	}
}

die "No words to use. Exiting.\n" if (scalar @selected_words <= 0);

@wordlist = ();
@selected_words = sort { length $a <=> length $b } @selected_words;

# Main loop
do {
	$completely_done = $num_of_words;
	$number_of_iterations=0;
	while ($completely_done > 0 && $number_of_iterations < $max_iterations) {
		print "Creating new puzzle $number_of_iterations\n" if $debug;
		$number_of_iterations++;
		$completely_done = $num_of_words;
		$success = 1;
		@new_word_list = @selected_words;
		&clear_grid;
		while (($new_word = pop @new_word_list) && $success) {
			$results = &place_word($new_word,$gridsize,$directions);
			if ($results == 0) {
				$completely_done--; 
			} else {
				print "iterating over grid\n" if $debug;
				$results = &iterate_word($new_word,$gridsize, $directions);
				if ($results != 0) {
					print "Grid iteration didn't work either\n" if $debug;
					$success = 0;
				} else {
					$completely_done--;
				}
			}
		}
	}
	print "Number of main loop iterations = $number_of_iterations\n" if $debug;
	if ($completely_done > 0) {
		if ($nogrow) {
			die "Can't create wordsearch with these parameters. Aborting.\n";
		} else {
			$gridsize+=5;
			warn "Increasing gridsize to $gridsize\n";
		}
	} else {
		$finished = 1;
	}
} until ($finished);

&print_grid(0);
print "\n\n";
@selected_words = sort @selected_words;
&print_words;

if (!$nosolution) {
	print "\n";
	&print_grid(1);
}

exit(0);


sub place_word {
	my ($word,$size,$possible_directions) = @_;
	my $done = 0;
	@letters = split //,$word;
	$iteration_counter = 0;

	while (!$done && (2*$num_of_words >= $iteration_counter) ) {
		$x = int (rand $size);
		$y = int (rand $size);

		if ($grid[$y][$x] eq ' ' || $grid[$y][$x] eq '') {
			$done = &test_word_position($x,$y,$size,$possible_directions);
		}

		$iteration_counter++;
	}
	if (!$done) {
		print "Random iterations failed: $iteration_counter\n" if $debug;
		return 1;
	}
	print "Random Iterations $iteration_counter\n" if $debug;
	return 0;
}

sub iterate_word {
	my ($word,$size,$possible_directions) = @_;
	my $done = 0;
	@letters = split //,$word;
	my $x=0;
	my $y=0;
	while (!$done && ($y<$size) ) {
		if ($grid[$y][$x] eq ' ' || $grid[$y][$x] eq '') {
			$done = &test_word_position($x,$y,$size,$possible_directions);
		}

		$x++;
		if ($x>=$size) {
			$x=0;
			$y++;
		}
	}
	if (!$done) {
		return 1;
	}
	return 0;
}
sub print_grid {
	my $solution = shift;
	for ($i=0;$i<$gridsize;$i++) {
		for ($j=0;$j<$gridsize;$j++) {
			if (($grid[$i][$j] eq '' || $grid[$i][$j] eq ' ') && (!$solution)) {
				print uc $randomletters[rand @randomletters];
			} else {
				print $grid[$i][$j];
			}
			print " ";
		}
		print "\n";
	}
}

sub print_words {
	# Copied from Perl Cookbook by Tom Christiansen & Nathan Torkington
	# Recipie 4.18
	my ($item, $cols, $rows, $maxlen);
	my ($xpixel, $ypixel, $mask, @data);
	$cols = 80;
	$maxlen = 1;        
	foreach $word (@selected_words) {
		my $mylen;
		$word =~ s/\s+$//;
		$maxlen = $mylen if (($mylen = length $word) > $maxlen);
		push(@data, $word);
	}

	$maxlen += 1;               # to make extra space

	# determine boundaries of screen
	$cols = int($cols / $maxlen) || 1;
	$rows = int(($#data+$cols) / $cols);

	# pre-create mask for faster computation
	$mask = sprintf("%%-%ds ", $maxlen-1);

	# subroutine to check whether at last item on line
	sub EOL { ($item+1) % $cols == 0 }  

	# now process each item, picking out proper piece for this position
	for ($item = 0; $item < $rows * $cols; $item++) {
		my $target =  ($item % $cols) * $rows + int($item/$cols);
		my $piece = sprintf($mask, $target < @data ? $data[$target] : "");
		$piece =~ s/\s+$// if EOL();  # don't blank-pad to EOL
			print $piece;
		print "\n" if EOL();
	}

	# finish up if needed
	print "\n" if EOL();
}

sub test_word_position {
	my $orig_x = shift;
	my $orig_y = shift;
	my $size = shift;
	my $possible_directions = shift;
	my $direction;
	$done = 0;
	@direction_choices = ();

	for ($direction = 0; $direction<$possible_directions; $direction++) {
		my $x = $orig_x;
		my $y = $orig_y;
		$error = 0;

		foreach $letter (@letters) {
			if ($x <= 0) {
				$x=0;
				$error = 1;
			}
			if ($x >= $size) {
				$x=$size;
				$error = 1;
			}
			if ($y <= 0) {
				$y=0;
				$error = 1;
			}
			if ($y >= $size) {
				$y=$size;
				$error = 1;
			}
			if (!(($grid[$y][$x] eq '') || ($grid[$y][$x] eq ' '))) { 
				if ($grid[$y][$x] ne $letter)  {
					$error = 1;
				}
			}
			$x+= $x_offset[$direction];
			$y+= $y_offset[$direction];

		}
		if (0 == $error) {
			push @direction_choices, $direction;
		} else {
			next;
		}
	}
	if (0 < scalar @direction_choices ) {
		print "Direction choices: " . scalar @direction_choices . "\n" if $debug;
		my $x = $orig_x;
		my $y = $orig_y;
		$rand_direction = @direction_choices[int(rand scalar @direction_choices)];
		print "$rand_direction\n" if $debug;

		foreach $letter (@letters) {

			if (($grid[$y][$x] eq '') || 
					($grid[$y][$x] eq ' ') || 
					($grid[$y][$x] eq $letter))  {
				$grid[$y][$x] = $letter;
				$x+= $x_offset[$rand_direction];
				$y+= $y_offset[$rand_direction];
			} else {
				die &print_grid;
			}
		}
		$done = 1;
	}
	return $done;
}

sub clear_grid{

	for ($i=0;$i<$gridsize;$i++) {
		for ($j=0;$j<$gridsize;$j++) {
			$grid[$i][$j] = ' ';
		}
	}
}

sub usage {
	print "Usage: $0 [OPTION] \n";
	print "Creates a word search puzzle.\n";
	print "\n";
	print " --size		Size of the grid (default=$gridsize).\n";
	print " --directions	Directions to place words (default=$directions).\n";
	print "		(Diagonals and reverse words = 8, No diagonals = 4,\n		 No reverse words = 2)\n";
	print " --words	Number of words to select (default=$num_of_words\n";
	print " --wordfile	Read words from a file instead of from default location.\n";
	print "		(Currently $wordfile)\n";
	print " --all		Use all words from the list of words provided.\n";
	print " 		(DO NOT USE THIS WITH THE DEFAULT WORD LIST LOCATION!)\n";
	print " --nosolution	Don't display the solution.\n";
	print " --nogrow	Don't grow the grid to find a solution.\n";
	print " --quick		Iterate one time through before trying new parameters.\n";
	print " --thorough	Iterate many more times through before trying new parameters.\n";
	print " --version	Display the version number.\n";
	print " --help		Display this help file.\n";
	print "\nReport bugs to <craig\@decafbad.net>\n"
}
